#!/usr/bin/perl
# HTCheckpoint
# One-shot HTTP server, with support for pausing and aborting transfer
# at arbitrary bytes.
# 
# Usage:
#     htcheckpoint [-p PORT] [-s HTTPSTATUS] [-t CONTENTTYPE]
#                  [-c CHECKPOINT...] [-P] [-d DELAY] [-A] [-v]
#                  URI FILENAME
# 
# Default port is 8085, default HTTP status code is 200, default content type
# is application/octet-stream.
# 
# 
# HTCheckpoint copyright (C) 2016 Nutchanon Wetchasit
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
use strict;
use Socket;
use Time::HiRes qw(usleep);

# Global variables
our $port = 8085;
our $uri;
our $filename;
our $contenttype = "application/octet-stream";
our $httpstatus = 200;
our $omitlength;
our @checkpoints;
our $pauseresponse;
our $delay;
our $abort = 0;
our $bufsize = 4096;
our $lineending = $/;
our $verbose = 0;

# parseopt(\@args)
# Parse command line parameters, exit with code 3 if failed, exit with code 0 if help/version page is requested
sub parseopt {
	my ($args) = @_;
	my ($i, $arg);
	my $acceptswitch;
	my $checkpoint;
	
	$acceptswitch = 1;
	for($i = 0; $i < scalar(@$args); $i++) {
		$arg = $args->[$i];
		if($acceptswitch and $arg eq "-A") {
			$abort = 1;
		} elsif($acceptswitch and $arg eq "-c") {
			if($i >= scalar(@$args) - 1) {
				print STDERR "htcheckpoint: option " . $arg . " requires a parameter\n";
				exit 1;
			}
			$checkpoint = $args->[$i + 1];
			if(not $checkpoint =~ /^\d+$/) {
				print STDERR "htcheckpoint: invalid checkpoint offset entered\n";
				exit 1;
			}
			push(@checkpoints, $checkpoint);
			$i++;
		} elsif($acceptswitch and $arg eq "-d") {
			if($i >= scalar(@$args) - 1) {
				print STDERR "htcheckpoint: option " . $arg . " requires a parameter\n";
				exit 1;
			}
			$delay = $args->[$i + 1];
			if(not $delay =~ /^\d+$/) {
				print STDERR "htcheckpoint: invalid delay time entered\n";
				exit 1;
			}
			$i++;
		} elsif($acceptswitch and $arg eq "-p") {
			if($i >= scalar(@$args) - 1) {
				print STDERR "htcheckpoint: option " . $arg . " requires a parameter\n";
				exit 1;
			}
			$port = $args->[$i + 1];
			if(not $port =~ /^\d+$/ or $port > 65535 or $port < 1) {
				print STDERR "htcheckpoint: invalid port number entered\n";
				exit 1;
			}
			$i++;
		} elsif($acceptswitch and $arg eq "-P") {
			$pauseresponse = 1;
		} elsif($acceptswitch and $arg eq "-h") {
			showhelp();
			exit 0;
		} elsif($acceptswitch and $arg eq "-L") {
			$omitlength = 1;
		} elsif($acceptswitch and $arg eq "-s") {
			if($i >= scalar(@$args) - 1) {
				print STDERR "htcheckpoint: option " . $arg . " requires a parameter\n";
				exit 1;
			}
			$httpstatus = $args->[$i + 1];
			if(not $httpstatus =~ /^\d\d\d$/) {
				print STDERR "htcheckpoint: invalid HTTP status code entered\n";
				exit 1;
			}
			if(not defined httpmessage($httpstatus)) {
				print STDERR "htcheckpoint: unrecogized HTTP status code " . $httpstatus . "\n";
				exit 1;
			}
			$i++;
		} elsif($acceptswitch and $arg eq "-t") {
			if($i >= scalar(@$args) - 1) {
				print STDERR "htcheckpoint: option " . $arg . " requires a parameter\n";
				exit 1;
			}
			$contenttype = $args->[$i + 1];
			$i++;
		} elsif($acceptswitch and $arg eq "-v") {
			$verbose = 1;
		} elsif($acceptswitch and $arg eq "-V") {
			showversion();
			exit 0;
		} elsif($acceptswitch and $arg eq "--") {
			$acceptswitch = 0;
		} elsif($acceptswitch and $arg =~ /^-/) {
			print STDERR "htcheckpoint: invalid option: " . $arg . "\n";
			exit 1;
		} else {
			if(not defined $uri) {
				$uri = $arg;
				next;
			} elsif(not defined $filename) {
				$filename = $arg;
				next;
			} else {
				print STDERR "htcheckpoint: too many parameters: " . $arg . "\n";
				exit 1;
			}
		}
	}
	
	if(not defined $uri or not defined $filename) {
		print STDERR "htcheckpoint: required parameters missing\n";
		exit 1;
	}
	@checkpoints = sort { $a <=> $b; } @checkpoints;
}

# httpmessage($status):$message
# Return HTTP status message according to status code
sub httpmessage {
	my ($status) = @_;
	my %messagemap = (
		200 => "OK",
		201 => "Created",
		204 => "No Content",
		301 => "Moved Permanently",
		302 => "Moved Temporarily",
		304 => "Not Modified",
		400 => "Bad Request",
		401 => "Unauthorized",
		403 => "Forbidden",
		404 => "Not Found",
		500 => "Internal Server Error",
		501 => "Not Implemented",
		502 => "Bad Gateway",
		503 => "Service Unavailable"
	);
	
	return $messagemap{$status};
}

# showhelp()
# Display help screen
sub showhelp() {
	print "Serve file via HTTP with pause on arbitrary position.\n";
	print "\n";
	print "htcheckpoint [OPTIONS] URI FILE\n";
	print "\n";
	print "  FILE           Filename to serve\n";
	print "  URI            Path that client access the file, e.g. \"/test.txt\"\n";
	print "\n";
	print "  -p PORT        Set TCP port number used by the server (default is 8085)\n";
	print "  -t MIMETYPE    Specify content type of the served file\n";
	print "                 (default is application/octet-stream)\n";
	print "  -s HTTPSTATUS  HTTP status code to use when serving file (default is 200)\n";
	print "  -L             Omit Content-Length field in HTTP response\n";
	print "  -c BYTE        Pause when number of transferred bytes reach specified value\n";
	print "                 (could be specified multiple times)\n";
	print "  -P             Pause before writing HTTP response\n";
	print "  -A             Abort transfer instead of pausing at the last checkpoint\n";
	print "                 specified by -c\n";
	print "  -d DELAY       Delay for specified miliseconds instead of pausing\n";
	print "  -v             Show verbose information on the operation\n";
	print "  -h             Show this help\n";
	print "  -V             Show version and author information\n";
}

# showversion()
# Display version screen
sub showversion() {
	print "HTCheckpoint 0.6.0\n";
	print "\n";
	print "Copyright (C) 2016 Nutchanon Wetchasit\n";
	print "License GPL: GNU GPL <http://www.gnu.org/licenses/gpl.txt>\n";
	print "This is free software: you are free to change and redistribute it.\n";
	print "There is NO WARRANTY, to the extent permitted by law.\n";
}

# Main program
my %getopts_list;

my $serversocket;
my $clientsocket;

my $i;
my $requestline;
my $location;
my $file;
my $length;
my $totalread;
my $buffer;
my $blocksize;
my $checkpoint;
my ($remaining, $tryread);
my $lastcheckpoint;

# Use unbuffered STDOUT
$|=1;

# Parse command line options
parseopt(\@ARGV);
if($verbose) {
	print "htcheckpoint: port: " . $port . "\n";
	print "htcheckpoint: filename: " . $filename . "\n";
	print "htcheckpoint: uri: " . $uri . "\n";
	print "htcheckpoint: content type: " . $contenttype . "\n";
}

# Open file, and check for size
if(not open($file, "<", $filename)) {
	print STDERR "htcheckpoint: failed to open file: " . $! . "\n";
	exit 3;
}
binmode($file);
$length = -s ($filename);
if(not defined $length) {
	print STDERR "htcheckpoint: file length chould not be checked: " . $! ."\n";
	close($file);
	exit 3;
}

# Remove duplicate checkpoints
for($i = 0; $i < scalar(@checkpoints) - 1; $i++) {
	if($checkpoints[$i] eq $checkpoints[$i + 1]) {
		splice(@checkpoints, $i, 1);
	}
}

if($verbose) {
	print "htcheckpoint: file length: " . $length . " byte(s)\n";
	if(scalar(@checkpoints) > 0) {
		print "htcheckpoint: checkpoint(s): " . join(",", @checkpoints) . "\n";
	}
	if($pauseresponse) {
		print "htcheckpoint: checkpoint before responding: yes\n";
	}
	if(defined $delay) {
		print "htcheckpoint: checkpoint delay: " . $delay . "\n";
	}
}

# Check for beyond-EOF checkpoints
for $checkpoint (@checkpoints) {
	if($checkpoint > $length) {
		print STDERR "htcheckpoint: warning: checkpoint position " . $checkpoint . " is beyond end of file\n";
	}
}

# Opening socket
if(not socket($serversocket, PF_INET, SOCK_STREAM, getprotobyname("tcp"))) {
	print STDERR "htcheckpoint: failed creating socket: " . $! . "\n";
	close($file);
	exit 3;
}
if(not setsockopt($serversocket, SOL_SOCKET, SO_REUSEADDR, 1)) {
	print STDERR "htcheckpoint: failed setting reusable flag on socket: " . $! . "\n";
	close($file);
	close($serversocket);
	exit 3;
}
if(not bind($serversocket, sockaddr_in($port, INADDR_ANY))) {
	print STDERR "htcheckpoint: failed binding port: " . $! . "\n";
	close($file);
	close($serversocket);
	exit 3;
}

# Listen for incomming connection
if(not listen($serversocket, 1)) {
	print STDERR "htcheckpoint: failed listening port: " . $! . "\n";
	close($file);
	close($serversocket);
	exit 3;
}
while(1) {
	if(not accept($clientsocket, $serversocket)) {
		if($!{EINTR}) {
			next;
		}
		print STDERR "htcheckpoint: failed recieving connection: " . $! . "\n";
		close($file);
		close($serversocket);
		exit(3);
	}
	last;
}
close($serversocket);

# Connected, prepare client socket
if($verbose) {
	print "htcheckpoint: client connected\n";
}
binmode($clientsocket);
select($clientsocket);
$| = 1;
select(STDOUT);
$/ = "\r\n";
$\ = undef;

# Read first request line
$requestline = <$clientsocket>;
chomp $requestline;
if($requestline =~ /^(\S+) (\S+) HTTP\/(\d+)\.(\d+)$/) {
	if($verbose) {
		print "htcheckpoint: request method: " . $1 . "\n";
	}
	if($1 ne "GET") {
		print $clientsocket "HTTP/1.0 501 Not Implemented\r\n";
		print $clientsocket "\r\n";
		close($file);
		close($clientsocket);
		exit 2;
	}
	$location = $2;
} else {
	print $clientsocket "HTTP/1.0 400 Bad Request\r\n";
	print $clientsocket "\r\n";
	close($file);
	close($clientsocket);
	exit 2;
}

# Read the rest of request headers
if($verbose) {
	print "htcheckpoint: requested uri: " . $location . "\n";
	print "htcheckpoint: reading request header fields\n";
}
while(1) {
	$requestline = <$clientsocket>;
	chomp $requestline;
	
	if($requestline eq "") {
		last;
	}
}

# Check URI
if($location ne $uri) {
	if($verbose) {
		print "htcheckpoint: uri not found\n";
	}
	print $clientsocket "HTTP/1.0 404 Not Found\r\n";
	print $clientsocket "\r\n";
	close($file);
	close($clientsocket);
	exit 2;
}

# Restore normal input line ending
$/ = $lineending;

if($pauseresponse) {
	# Checkpoint before response (when specified)
	if(scalar(@checkpoints) == 0) {
		$lastcheckpoint = 1;
	}
	
	if($lastcheckpoint and $abort) {
		# Abort transfer
		print "htcheckpoint: request received, aborting connection\n";
		close($file);
		close($clientsocket);
		exit 0;
	}
	
	if(not defined $delay) {
		print "htcheckpoint: request received, press ENTER to continue\n";
		<STDIN>;
	} else {
		print "htcheckpoint: request received, pausing " . $delay ." ms\n";
		usleep($delay * 1000);
	}
}

# Write response header
print $clientsocket "HTTP/1.0 " . $httpstatus . " " . httpmessage($httpstatus) . "\r\n";
print $clientsocket "Content-Type: " . $contenttype . "\r\n";
if(not $omitlength) {
	print $clientsocket "Content-Length: " . $length . "\r\n";
}
print $clientsocket "\r\n";
if($verbose) {
	print "htcheckpoint: response header sent\n";
}

# Write response body
$totalread = 0;
undef $checkpoint;
while(1) {
	# Checkpoint action
	if(defined $checkpoint) {
		# Hit a target, do specified action
		
		if($checkpoint < 0) {
			# End of file
			last;
		}
		
		if($lastcheckpoint and $abort) {
			# Abort transfer
			print "htcheckpoint: sent " . $totalread . " byte(s), aborting connection\n";
			close($file);
			close($clientsocket);
			exit 0;
		}
		
		if(not defined $delay) {
			print "htcheckpoint: sent " . $totalread . " byte(s), press ENTER to continue\n";
			<STDIN>;
		} else {
			print "htcheckpoint: sent " . $totalread . " byte(s), pausing " . $delay ." ms\n";
			usleep($delay * 1000);
		}
	}
	
	# Find next checkpoint
	while(1) {
		$checkpoint = shift(@checkpoints);
		if(not defined $checkpoint) {
			# No more checkpoint, set target to EOF
			$checkpoint = -1;
			last;
		}
		
		if($totalread <= $checkpoint) {
			# Next non-passed checkpoint found
			if($checkpoint > $length) {
				# Next checkpoint is beyond EOF, set target to EOF
				$checkpoint = -1;
			}
			last;
		}
	};
	if(scalar(@checkpoints) == 0) {
		$lastcheckpoint = 1;
	}
	
	# Transfer until target reached
	while(1) {
		if($checkpoint >= 0) {
			$remaining = $checkpoint - $totalread;
			if($verbose) {
				print "htcheckpoint: next checkpoint " . $checkpoint . " current " . $totalread . " remaining " .$remaining . "\n";
			}
		} else {
			$remaining = $length - $totalread;
			if($verbose) {
				print "htcheckpoint: read to the end, current " . $totalread . " remaining " .$remaining . "\n";
			}
		}
		
		if($remaining <= 0) {
			# Reached current target
			last;
		}
		
		$tryread = $remaining < $bufsize ? $remaining : $bufsize;
		if($verbose) {
			print "htcheckpoint: try reading " . $tryread . "\n";
		}
		
		$blocksize = read($file, $buffer, $tryread);
		if($verbose) {
			print "htcheckpoint: read " . $blocksize . "\n";
		}
		if(not defined $blocksize) {
			print STDERR "htcheckpoint: failed reading file: " . $! . "\n";
			close($file);
			close($clientsocket);
			exit 3;
		} elsif($blocksize == 0) {
			print STDERR "htcheckpoint: unexpected end of file\n";
			close($file);
			close($clientsocket);
			exit 2;
		}
		print $clientsocket $buffer;
		
		$totalread += $blocksize;
	}
}

if($verbose) {
	print "htcheckpoint: closing connection\n";
}
close($file);
close($clientsocket);
